/* 
    psgl.c copyright (c) 1990, 1991, 1992  seth j. teller, seth@miro.berkeley.edu.

    permission is granted to modify and redistribute this code in any manner
    as long as this notice is preserved.  all standard disclaimers apply.
    
    mod log:
	spring   90:	original module written 
	fall     91:	fixed order of bgn() feedback calls as per bernhard hofmann's suggestions
	fall     91:	added *qstrip() calls 
	november 91:	fixed bug in color summation reported by Karsten Gaier (gaier@tat.physik.uni-tuebingen.de)
	winter   91:	static'ed all internal vars & routines, to limit namespace pollution
	spring   92:	documentation clarified
*/

#include "math.h"
#include "stdio.h"
#include "device.h"
#include <gl/gl.h>
#include <gl/feed.h>
#include <gl/get.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <strings.h>

/* n.b.: one machine #define should appear _before_ "psgl.h" in your source file */
/* or, equivalently, use -Dxxx while compiling your C files, including psgl.c */
/* #define SGIHW_PI  */
/* #define SGIHW_GTX */
#define SGIHW_VGX

/* next three lines required */
#ifndef PSGL
#define PSGL
#endif
#define PSGLCFILE
#include "psgl.h"

/*
    psgl.c == subsetted postscript gl.

    this module converts gl calls to near-wysiwyg color postscript.

	author seth teller, seth@miro.berkeley.edu/seth@sgi.com, march 1990.

    to use psgl, include the following 3 lines in all your C files that make gl calls
    (equivalently, include these three lines in some .h file that all your sources include):

    #ifdef PSGL
    #include "psgl.h"
    #endif

    and compile (i.e., cc) -DPSGL to enable the psgl routines.

    note: one of -DSGIHW_PI, -DSGIHW_GTX, -DSGIHW_VGX is necessary to specify platform;
    or, you could #define this above the #include "psgl.h" in your C file

    finally, compile this routine as you would any normal c file 
    (say, with cc -c psgl.c), include psgl.o in the list of .o files 
    given to the loader, and go.

    anyone using this file should #include "psgl.h" _after_ including "gl.h".

    this file replaces the following gl calls:

	name						    comments

	void Pick()
	void Endpick()
        void Perspective()
	void Ortho()
        void Ortho2()
        void Viewport()
	void Popviewport()
	void Setdepth()
	void Lsetdepth()
	void Reshapeviewport()
	void Cpack()
	void rgbcolor()					    note lower case rgb
	void Arc*()
	void Arcf*()
	void Circ*()
	void Circf*()
	void Crv()
	void Crvn()
	void Rcrv()
	void Rcrvn()
	void Cmov()					    getcpos() doesn't work
	void Charstr()
	void Pnt()
	void Draw()
	void Rdr()
	void Rect*()
	void Rectf*()
        void Polf()
        void Polfi()
        void Polfs()
	void Polf2()
	void Polf2f()
	void Polf2s()
	void Bgnpolygon()
	void Endpolygon()
	void Bgnline()
	void Endline()
	void Bgnpoint()
	void Endpoint()
	void Bgnclosedline()
	void Endclosedline()
	void Bgntmesh()
	void Endtmesh()
	void Bgnqstrip()
	void Endqstrip()
	void Swapbuffers()				    writes file 'psgl_fname'
	void Zbuffer()
	void Backface()
	long Winopen()					    returns a long winid
	long Qread()
	void Bgnsurface()
	void Endsurface()
	void Clear()
	void Backbuffer()
	void Frontbuffer()

    note that these routines are capitalized differently than they appear in the gl.
    including "psgl.h" will effect this change in your code.

    the set of calls above are replaced with local routines that cache the graphics 
    objects written, in 2-d coordinates presumably suitable for a standard postscript 
    device (here, an 8 1/2" x 11" laserwriter).

    the output is written in EPS (encapsulated postscript) format (see adobe's ps
    ref books for details).

    specifically: the cache object types defined are

	CACHED_POINT		    treated as zero-length line
	CACHED_VECTOR		    move/draw sequence; bgn/endline sequence
	CACHED_VECTORLOOP	    bgn/endclosedline sequence
	CACHED_POLYGON		    bgn(), v3f(), .... end()
	CACHED_STRING		    charstr()

    they are transformed to screen (x,y,depth) coordinates
    and stored in generation order.  at swapbuffers() time,
    the objects are crudely sorted according to a depth,
    then output to a file.

    the psgl module intercepts all calls to, and returns from, qread().
    this is obviously going to cause a performance hit, probably comparable
    to that taken when using the shared graphics library instead of the
    machine-dependent version.

    pressing leftctrl-shift-printscrn or rightctrl-shift-printscrn
    prompts the psgl module to queue a REDRAW event.  the client
    program, upon receiving the REDRAW, should handle it appropriately;
    that is, by refreshing the display.  at this point, the psgl code
    is enabled and will intercept the relevant gl calls up to and
    including the next call to

	swapbuffers().

    each drawn object will be stored in screen coordinates.  at
    swap time, the objects are sorted by depth and output back to
    front in postscript format.

    some things to note:

	0) you can call the psgl routines directly, if you wish.  
	    call psgl_open (fname) with string fname as the name of a 
	    writeable file. calling it with NULL or "" implies the name 
	    "windowname.n.psgl", where n is omitted when zero and included
	    for n > 0, and where 'windowname' is the string passed to winopen().

	1) the current window is mapped to a vertical 8x10 postscript page,
	    _preserving_ aspect ratio.

	2) an EPS (encapsulated postscript) %%BoundingBox line
            is written for use of dvi2ps, latex, and other programs.
	    the resulting document has a single page only.

	3) characters are treated as objects with depth, as is everything else.

	4) calling swapbuffers() writes _one_ file and reverts to normal gl mode.

	5) color is handled by reading r,g,b from feedback mode.  setrgbcolor
	    postscript ops are issued; these are handled okay by b&w laserwriters.

	6) vectors and vectorloops are handled separately to take
	    advantage of postscript mitreing of line joins.

	7) all postscript is clipped to the bounding box
	    of drawn objects; this bbox is clipped to the graphics window.

	8) the *i and *s routines (integer and short) are handled as
	    redirections to the floating-point calls (via macros in psgl.h).

    context-ful operations should not be intermixed.  for example,
    charstr()'s should not occur inside bgn*()..end*() pairs, etc.
    this is analogous to gl-type restrictions, but somewhat more demanding.

    if the z sorting does stupid things, the user program
    should output smaller objects (perhaps with adaptive subdivision).
    i've written a psgl lib that subdivides adaptively in screen space,
    and is somewhat of a hack, but i'd be glad to make it available.

    some further notes:

	some subtle changes may occur in program behavior, since this
    module queues and intercepts events due to particular devices (see below)
    
	this module calls malloc() and realloc(), but once it govs up
    to your problem size there should be no more allocations.

	the only externally visible symbols should be the modified
    gl calls enumerated above (of course), and a few others such
    as psgl_open().  all others are static'ed to be module-private.

	the gl has a platform-dependent hard limit on the amount of stuff
    allowed in the feedback buffer.  psgl will print an error message if
    it detects an overflow.  there's no nice workaround to this; you may
    have to bite the bullet and make your problem size smaller.  the only
    time this has hurt me is while outputting NURBs with tens of thousands
    of polygons.

	showcase output mode is not yet implemented

	indigo and crimson feedback modes are not yet implemented
*/

#define CACHED_POINT	    1
#define CACHED_VECTOR	    2
#define CACHED_VECTORLOOP   3
#define CACHED_POLYGON	    4
#define CACHED_STRING	    5

typedef struct pointstruct {
    float px;
    float py;
    float pz;
    } POINT;

struct grfix_object {
    short type;		/* just CACHED_VECTOR, _POLYGON, _STRING    */
    short length;	/* number of components; type dependent	    */
    long color;		/* alpha, b, g, r			    */
    char *data;		/* xformed 3d data; strings; etc.	    */
    float x, y, z;	/* reference point for this object	    */
    float w, h;		/* width and height if needed		    */
    float   xmin,ymin,
	    xmax,ymax,
	    zmin,zmax;  /* object bounding box			    */
    int timestamp;	/* unique drawing time, zero-relative	    */
    short descriptor;	/* norgb, offwindow flags, zbuffer, etc.    */
    };

#define PSGL_NORMAL	0x0
#define PSGL_NOCOLOR	0x1	    /* couldn't find pixel on object	*/
#define PSGL_OFFWINDOW	0x2	    /* bbox disjoint from window	*/
#define PSGL_ZBUFFER	0x4	    /* state at time object was drawn	*/

typedef struct grfix_object CACHE_ITEM;

#define PSGL_POSTSCRIPT 1
#define PSGL_SHOWCASE 2

static	int psgl_mode = PSGL_POSTSCRIPT;

static	CACHE_ITEM	*psgl_cache;		    /* (subject to realloc) */
static	int		psgl_cache_index = 0;	    /* next free item */
static	int		psgl_max_items = 256;	    /* current max index (dynamic) */
static	int		psgl_max_bytes = 8 * 1024;  /* current max data (dynamic)  */
static	char		*psgl_cache_base;
static	char		*psgl_cache_data;

static	FB_TOKEN	*psgl_fbbuff = NULL;	    /* for feedback results    */

#if defined(PI) || defined(VGX)
static	int		psgl_max_fback = 8 * 1024;  /* max # of token/data shorts returned */
#else	/* GTX */
static	int		psgl_max_fback = 32 * 1024; /* max # of token/data shorts returned */
#endif

static	int		psgl_cur_fback = 0;	    /* current # of valid tokens    */
static	int		psgl_fb_omitted = FALSE;    /* did we possibly lose fb data?*/
static	int		psgl_equal_z = 0;	    /* how many collisions in sort? */

static	float		psgl_cx, psgl_cy, psgl_cz;  /* our idea of cpos(), charstr() motion */
static	int		psgl_ortho = FALSE;	    /* TRUE iff ortho() or ortho2() call is in effect */
static	int		psgl_zbuffer = FALSE;	    /* zbuffer shadow state */
static	long		psgl_bufferstate;	    /* front & back buffers at entry */
static	long		psgl_backbuffer = TRUE;	    /* user's idea of back buffer state */
static	long		psgl_frontbuffer;	    /* user's idea of front buffer state */
static	long		psgl_zbufferstate = FALSE;  /* zbuffer state at entry */
static	long		psgl_backfacestate = FALSE; /* backface state at entry */
static	long		psgl_userbackface = FALSE;  /* dump with or w/out backface culling? */
static	int		psgl_timestamp = 0;	    /* total order on drawn objects */

static	short		psgl_cposx, psgl_cposy;	    /* receivers of getcpos() calls -- unused */

static	Screencoord	psgl_left, psgl_right, 
			psgl_bottom, psgl_top;	    /* viewport */
static	long		psgl_zbnear, psgl_zbfar;    /* zbuffer range */
static	float		psgl_cpnear, psgl_cpfar;    /* front & back planes */
static	long		psgl_xorg, psgl_yorg,
			psgl_xsize, psgl_ysize;	    /* window */
static	float		psgl_xmin, psgl_ymin,
			psgl_xmax, psgl_ymax;	    /* bounding box */
static	float		psgl_w, psgl_h;		    /* bounding extents */

static	long		psgl_color = 0L;	    /* our idea of current cpack() or RGBcolor() target */
static	char		psgl_fname[128], 
			psgl_prefix[128];
static	char		*psgl_epilogue;

static	FILE		*psgl_file = NULL;	    /* file for cache dump */
static	int		psgl_triggered = FALSE;	    /* this flag is set by interrupt handler, cleared by Swapbuffers() */

						    /* N.B. psgl_enabled, psgl_debug_level exported */
int	    		psgl_enabled = FALSE;	    /* this flag is set by psgl_open(), cleared by Swapbuffers() */
int			psgl_debug_level = 0;	    /* increasing this yields verbosity, ps bounding boxes, etc. */

#define MAX_GLVERTICES 256			    /* max v's inside bgn*() .. end*() pair */
static	int		psgl_vnum = 0;		    /* number of vertices currently held    */
static	POINT		psgl_vbuff[MAX_GLVERTICES]; /* for temporary storage of gl vertices */

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))
#define AVG(a,b) (((a) + (b)) * 0.5)
#define INTERVAL(a,b,c) ((a) <= (b) && (b) <= (c))

/* exported */
char *psgl_open(), *psgl_open_internal();

/* forward decl */
static int
insert_cache (
    int type,
    char *data,
    int n);

/* psgl_open returns the name of the file to be written, or NULL if error condition */
/* exported */
char *
psgl_open(
    char *fname)
{
    return (psgl_open_internal (fname, "showpage\n"));
}

static void
psgl_restore_client_state()
{
    /* restore client state */
    zbuffer (psgl_zbufferstate);
    backface (psgl_backfacestate);

    /* restore buffer state */
    if (psgl_bufferstate != DRAWZBUFFER)    {	    /* what if it is? */
	backbuffer  (psgl_bufferstate & 0x1);
	frontbuffer (psgl_bufferstate & 0x2);
	}
}

/* returns the name of the file to be written, or NULL if error condition */
/* exported */
char *
psgl_open_internal (
    char *fname,
    char *epilogue)
{
static ncalls = 0;
int k, slen;

    if (psgl_file != NULL)  {
	printf ("psgl.psgl_open_internal(): non-matched call?\n");
	fclose (psgl_file);
	psgl_file = NULL;
	psgl_restore_client_state();
	}

    if (psgl_epilogue != NULL)	{
	free (psgl_epilogue);
	psgl_epilogue = NULL;
	}

    if (epilogue != NULL && (slen = strlen(epilogue)) > 0) {
	psgl_epilogue = (char *)malloc (slen + 1);
        strncpy (psgl_epilogue, epilogue, slen + 1);
	}
    else psgl_epilogue = NULL;

    if (fname != NULL && strlen (fname) > 0)    /* use user supplied name */
        strcpy (psgl_fname, fname);
    else    {
	char prefix[40];

	if (isalpha(psgl_prefix[0]))
	    strcpy (prefix, psgl_prefix);
	else
	    sprintf (prefix, "%d", getpid() % 10000);

	if (psgl_mode == PSGL_POSTSCRIPT)   {
	    if (ncalls == 0)
		sprintf (psgl_fname, "%s.psgl", prefix);
	    else
		sprintf (psgl_fname, "%s.%d.psgl", prefix, ncalls);
	    }
	else	{
	    if (ncalls == 0)
		sprintf (psgl_fname, "%s.grn", prefix);
	    else
		sprintf (psgl_fname, "%s.%d.grn", prefix, ncalls);
	    }
	}
    ncalls++;

    slen = strlen (psgl_fname);
    for (k = 0; k < slen; k++)
	if (psgl_fname[k] == ' ')
	    psgl_fname[k] = '_';

    if ((psgl_file = fopen (psgl_fname, "w")) == NULL)	{
	printf ("psgl.psgl_open_internal(): can't open postscript file '%s'\n", psgl_fname);
	return (NULL);
	}

    if (psgl_cache == NULL)
    	psgl_cache = (CACHE_ITEM *)malloc (psgl_max_items * sizeof (CACHE_ITEM));

    if (psgl_cache_base == NULL)
	psgl_cache_base = (char *)malloc (psgl_max_bytes);

    if (psgl_cache == NULL || psgl_cache_base == NULL)   {
	fprintf (stderr,"psgl.psgl_open_internal(): can't malloc cache structure/data\n");
	return (NULL);
	}

    if (psgl_fbbuff == NULL)
    	psgl_fbbuff = (FB_TOKEN *)malloc (psgl_max_fback * sizeof (FB_TOKEN));

    if (psgl_fbbuff == NULL)	{
	fprintf (stderr,"psgl.psgl_open_internal(): can't malloc feedback buffer\n");
	return (NULL);
	}

    fprintf (stderr,"psgl.psgl_open_internal(): opened postscript file '%s'\n", psgl_fname);

    psgl_cache_index = 0;
    psgl_cache_data = psgl_cache_base;
    psgl_cache[0].data = NULL;
    psgl_fb_omitted = FALSE;
    psgl_equal_z = 0;

    /* get window attributes (pixels) */
    getorigin (&psgl_xorg, &psgl_yorg);
    getsize   (&psgl_xsize, &psgl_ysize);

    /* get viewport w,h */
    getviewport (&psgl_left, &psgl_right, &psgl_bottom, &psgl_top);

    /* initialize depth to something reasonable */
    psgl_zbnear = 0L;
    psgl_zbfar  = 0x7fffffL;

    /* force client to redraw ONCE; this is the draw we'll catch */
    qenter (REDRAW, 1);

    psgl_backfacestate = getbackface();
    psgl_bufferstate = getbuffer();
    psgl_zbufferstate = psgl_zbuffer = getzbuffer();

    backface(psgl_userbackface);	/* set/cleared by psgl_trigger code */
    zbuffer (FALSE);
    frontbuffer (TRUE);			/* draw to front */
    backbuffer (TRUE);	

    psgl_enabled = TRUE;
    psgl_timestamp = 0;

    /* seed random # generator */
    srand48 (getpid());

    return (psgl_fname);
}

static void
psgl_install()
{
static long psgl_installed = FALSE;

    if (!psgl_installed) {
	psgl_installed = TRUE;
        /* if necessary, queue L & R ctrl keys, shift keys, single SCREENPRINTKEY */
	if (!isqueued(LEFTCTRLKEY))
	    qdevice (LEFTCTRLKEY);
	if (!isqueued(RIGHTCTRLKEY))
	    qdevice (RIGHTCTRLKEY);
	if (!isqueued(LEFTSHIFTKEY))
	    qdevice (LEFTSHIFTKEY);
	if (!isqueued(RIGHTSHIFTKEY))
	    qdevice (RIGHTSHIFTKEY);
	if (!isqueued(PRINTSCREENKEY))
	    qdevice (PRINTSCREENKEY);
	if (!isqueued(GKEY))
	    qdevice (GKEY);
	if (!isqueued(PKEY))
	    qdevice (PKEY);

        fprintf (stderr, "psgl: (left)ctrl-shift-PrintScrn for postscript %s.*.psgl (no backface culling).\n", psgl_prefix);
        fprintf (stderr, "psgl: (right)ctrl-shift-PrintScrn for postscript %s.*.psgl (w/ backface culling).\n", psgl_prefix);
	}
}

/* only compares x & y */
static int
disjoint_cache_bbox (
    CACHE_ITEM *a, CACHE_ITEM *b)
{
    return (a->xmax < b->xmin || a->ymax < b->ymin || a->xmin > b->xmax || a->ymin > b->ymax);
}

/* cicomp (A,B): compare z depth of two cache items *A, *B. */
/* if zbuffering was off either or both of A and B, later one wins */
/* otherwise, compare z depths: */
/* A farther than B: -1 */
/* A closer than B:   1 */
/* farthest items end up at beginning of list, closest near start. */
static int 
cicomp (
    CACHE_ITEM *A, 
    CACHE_ITEM *B)
{
int Azb = A->descriptor & PSGL_ZBUFFER;
int Bzb = B->descriptor & PSGL_ZBUFFER;

    if (!(Azb && Bzb))	{
	if (disjoint_cache_bbox(A,B))	{
	    if (!Azb && Bzb)
		return (-1);
	    else if (Azb && !Bzb)
		return (1);
	    }
	/* either not disjoint, or neither was zbuffered */
	if (A->timestamp > B->timestamp)
	    return (1);				/* A was drawn later; it wins */
	else
	    return (-1);			/* B was drawn later; it wins */
	}

    if (A->z > B->z)	    /* compare depths; we're looking towards -zinf */
	return (-1);	    /* A is farther; draw it earlier */
    else if (A->z < B->z)
	return (1);	    /* A is closer; draw it later */
    else    {
	psgl_equal_z++;

	if (A->timestamp > B->timestamp)
	    return (1);				/* A was drawn later; it wins */
	else
	    return (-1);			/* B was drawn later; it wins */

#ifdef ZTIE
        if (A->zmin > B->zmin)		/* A's closest is farther than B's*/
	    return (-1);
	else if (A->zmin < B->zmin)
	    return ( 1);
	else
	    return (0);
#endif
	}
}

static void
init_cache_bbox (
    CACHE_ITEM *c)
{
    c->xmin =  1E20;
    c->ymin =  1E20;
    c->zmin =  1E20;

    c->xmax = -1E20;
    c->ymax = -1E20;
    c->zmax = -1E20;
}

static void
insert_cache_bbox (
    CACHE_ITEM *c,
    float x, float y, float z)
{
    c->xmin = MIN (x, c->xmin);
    c->xmax = MAX (x, c->xmax);
    c->ymin = MIN (y, c->ymin);
    c->ymax = MAX (y, c->ymax);
    c->zmin = MIN (z, c->zmin);
    c->zmax = MAX (z, c->zmax);
}

static int
bbox_off_window(
    CACHE_ITEM *a)
{
    return (a->xmax < 0.0 || a->ymax < 0.0 || a->xmin > psgl_xsize || a->ymin > psgl_ysize);
}

/* uses globals xmin,ymin,xmax,ymax */
static void
insert_bbox (
    float x, float y)
{
    psgl_xmin = MIN (x, psgl_xmin);
    psgl_xmax = MAX (x, psgl_xmax);
    psgl_ymin = MIN (y, psgl_ymin);
    psgl_ymax = MAX (y, psgl_ymax);
}

/* start feedback */
static void
open_feedback()
{
    feedback (psgl_fbbuff, psgl_max_fback);
}

/* grab feedback */
static int
close_feedback()
{
    psgl_cur_fback = endfeedback (psgl_fbbuff);

    if (psgl_cur_fback >= psgl_max_fback)   {	    /* may have lost data */
	psgl_fb_omitted = TRUE;
    	psgl_fbbuff = (FB_TOKEN *)realloc (psgl_fbbuff, (psgl_max_fback *= 2) * sizeof (FB_TOKEN));

	if (psgl_fbbuff == NULL)  {
	    fprintf (stderr,"psgl.close_feedback(): can't realloc() %d bytes for feedback buffer\n", 
		 psgl_max_fback * sizeof (short));
	    exit (-1);
	    }
	fprintf (stderr,"psgl.close_feedback(): max feedback tokens now %d (%d bytes)\n",
		    psgl_max_fback, psgl_max_fback * sizeof (short));
	}

    return (psgl_cur_fback);
}

#ifdef SGIHW_VGX

/* VGX */
/* parse a gl feedback buffer, and insert contents into cache */
static void
insert_feedback_cache (
    float *buff,
    long n)
{
int k = 0, i;
short type, count, x, y;
long r,g,b,a;
long z;
float s,t;

    while (k < n)   {
        psgl_vnum = 0;

	type = (short)buff[k++];
	count = (short)buff[k++];
       if (type == (short)FB_CMOV) {
		x = (short)buff[k++];
		y = (short)buff[k++];
		z = (long)buff[k++];

		r = (long)buff[k++];
		g = (long)buff[k++];
		b = (long)buff[k++];
		a = (long)buff[k++];

		s = buff[k++];
		t = buff[k++];

		psgl_cx = (float)(x - psgl_xorg);
		psgl_cy = (float)(y - psgl_yorg);
		psgl_cz = (float)(z);
	    }
	else if (type == (short)FB_PASSTHROUGH)	{
		k += 1;
	    }
	else if (type == (short)FB_POINT || type == (short)FB_LINE || type == (short)FB_POLYGON)    {
		r = g = b = a = 0L;
	    	for (i = 0; i < count / 9; i++) {
		    x = (short)buff[k++];
		    y = (short)buff[k++];
		    z = (long)buff[k++];

		    r += (long)buff[k++];
		    g += (long)buff[k++];
		    b += (long)buff[k++];
		    a += (long)buff[k++];

		    s = buff[k++];
		    t = buff[k++];

		    /* convert to floating point, zero-relative window coords */
		    psgl_vbuff[psgl_vnum  ].px = (float)(x - psgl_xorg);
		    psgl_vbuff[psgl_vnum  ].py = (float)(y - psgl_yorg);
    		    psgl_vbuff[psgl_vnum++].pz = (float)(z);
		    }	/* for i */

		/* use average of all vertex colors */
		b /= (count / 9);
		g /= (count / 9);
		r /= (count / 9);
		a /= (count / 9);

	        psgl_color = ((b & 0x000000ff) << 16) | ((g & 0x000000ff) << 8) | (r & 0x000000ff);
		insert_cache (type == FB_POINT	? CACHED_POINT
				: type == FB_LINE ? CACHED_VECTOR
				: /* type == FB_POLYGON ? */ CACHED_POLYGON, (char *)psgl_vbuff, count / 9);
		}
	    else    {
		fprintf (stderr,"psgl.insert_fb_cache(): garbage type %d?\n", type);
		return;
		}
	}   /* while k < n */
}   /* insert_f_c() */

#elif SGIHW_PI

/* PERSONAL IRIS */
/* GT graphics library user's guide, section 17.2 */
/* parse a gl feedback buffer, and insert contents into cache */
static int
insert_feedback_cache (
    float *buff,
    long n)
{
int k = 0, i;
short type, count, x, y;
long r,g,b;
long z;

    while (k < n)   {
	type = (short)buff[k++];
	if (type != (short)FB_DRAW && type != (short)FB_PASSTHROUGH)
	    psgl_vnum = 0;

	if (type == (short)FB_CMOV) {
	    x = (short)buff[k++];
	    y = (short)buff[k++];
	    z = (long)buff[k++];

	    psgl_cx = (float)(x - psgl_xorg);
	    psgl_cy = (float)(y - psgl_yorg);
	    psgl_cz = (float)(z);
	    }
	else if (type == (short)FB_PASSTHROUGH)	{
	    k++;
	    }
	else if (type == (short)FB_MOVE)    {
		r = g = b = 0L;
		count = 6;
	    	for (i = 0; i < count / 6; i++) {
		    x = (short)buff[k++];
		    y = (short)buff[k++];
		    z = (long)buff[k++];

		    r = (long)buff[k++];
		    g = (long)buff[k++];
		    b = (long)buff[k++];

		    /* convert to floating point, zero-relative window coords */
		    psgl_vbuff[psgl_vnum  ].px = (float)(x - psgl_xorg);
		    psgl_vbuff[psgl_vnum  ].py = (float)(y - psgl_yorg);
    		    psgl_vbuff[psgl_vnum++].pz = (float)(z);
		    }	/* for i */

		/* use average of all vertex colors */
		b /= (count / 6);
		g /= (count / 6);
		r /= (count / 6);
	    }
	else if (type == (short)FB_DRAW)    {
		r = g = b = 0L;  /* zeroed out above? */
		count = 6;
	    	for (i = 0; i < count / 6; i++) {
		    x = (short)buff[k++];
		    y = (short)buff[k++];
		    z = (long)buff[k++];

		    r = (long)buff[k++];
		    g = (long)buff[k++];
		    b = (long)buff[k++];

		    /* convert to floating point, zero-relative window coords */
		    psgl_vbuff[psgl_vnum  ].px = (float)(x - psgl_xorg);
		    psgl_vbuff[psgl_vnum  ].py = (float)(y - psgl_yorg);
    		    psgl_vbuff[psgl_vnum++].pz = (float)(z);
		    }	/* for i */

		/* use average of all vertex colors */
		b /= (count / 6);
		g /= (count / 6);
		r /= (count / 6);

	        psgl_color = ((b & 0x000000ff) << 16) | ((g & 0x000000ff) << 8) | (r & 0x000000ff);
		if (psgl_vnum > 1)
		    insert_cache (CACHED_VECTOR, (char *)&psgl_vbuff[psgl_vnum - 2], 2);
	    }
	else if (type == (short)FB_POINT || type == (short)FB_POLYGON)  {
		count = (type == (short)FB_POLYGON) ? (short)buff[k++] : 6;
		r = g = b = 0L;
	    	for (i = 0; i < count / 6; i++) {
		    x = (short)buff[k++];
		    y = (short)buff[k++];
		    z = (long)buff[k++];

		    r += (long)buff[k++];
		    g += (long)buff[k++];
		    b += (long)buff[k++];

		    /* convert to floating point, zero-relative window coords */
		    psgl_vbuff[psgl_vnum  ].px = (float)(x - psgl_xorg);
		    psgl_vbuff[psgl_vnum  ].py = (float)(y - psgl_yorg);
    		    psgl_vbuff[psgl_vnum++].pz = (float)(z);
		    }	/* for i */

		/* use average of all vertex colors */
		b /= (count / 6);
		g /= (count / 6);
		r /= (count / 6);

	        psgl_color = ((b & 0x000000ff) << 16) | ((g & 0x000000ff) << 8) | (r & 0x000000ff);
		insert_cache (type == FB_POINT ? CACHED_POINT : CACHED_POLYGON, (char *)psgl_vbuff, count / 6);
	    }
	else if (type == (short)FB_ZBUFFER || type == (short)FB_LINESTYLE || type == (short)FB_SETPATTERN
		|| type == (short)FB_LINEWIDTH || type == (short)FB_LSREPEAT)	{
	    k++;
	    }
        else    {
	    fprintf (stderr, "psgl.insert_fb_cache(): garbage type %d?\n", type);
	    return;
	    }
	}   /* while k < n */
}   /* insert_f_c() */

#elif SGIHW_GTX

/* GTX */
/* parse a gl feedback buffer, and insert contents into cache */
static void
insert_feedback_cache (
    short *buff,
    long n)
{
int k = 0, i;
short type, count, x, y;
long r,g,b,a;
long z;
long zhi, zlo;

    while (k < n)   {
        psgl_vnum = 0;

	type = buff[k++];
	count = buff[k++];
        switch ((int)type)   {
	    case (int)FB_CMOV :
		x = buff[k++];
		y = buff[k++] - 1024;
		zhi = buff[k++];
		zlo = buff[k++];
		z = (zhi << 16) + zlo;
		k += 4;

		psgl_cx = (float)(x - psgl_xorg);
		psgl_cy = (float)(y - psgl_yorg);
		psgl_cz = (float)(z);
		break;
	    
	    case (int)FB_PASSTHROUGH : 
		k += 8;		    /* ignore for now */
		break;

	    case (int)FB_POINT :
	    case (int)FB_LINE  :
	    case (int)FB_POLYGON :
		r = g = b = a = 0L;
	    	for (i = 0; i < count / 8; i++) {
		    x = buff[k++];
		    y = buff[k++] - 1024;
		    zhi = buff[k++];
		    zlo = buff[k++];
		    z = (zhi << 16) + zlo;
		    r += buff[k++];
		    g += buff[k++];
		    b += buff[k++];
		    a += buff[k++];

		    /* convert to floating point, zero-relative window coords */
		    psgl_vbuff[psgl_vnum  ].px = (float)(x - psgl_xorg);
		    psgl_vbuff[psgl_vnum  ].py = (float)(y - psgl_yorg);
    		    psgl_vbuff[psgl_vnum++].pz = (float)(z);
		    }	/* for i */

		/* use average of all vertex colors */
		b /= (count / 8);
		g /= (count / 8);
		r /= (count / 8);
		a /= (count / 8);

	        psgl_color = ((b & 0x000000ff) << 16) | ((g & 0x000000ff) << 8) | (r & 0x000000ff);
		insert_cache (type == FB_POINT	? CACHED_POINT
				: type == FB_LINE ? CACHED_VECTOR
				: /* type == FB_POLYGON ? */ CACHED_POLYGON, (char *)psgl_vbuff, count / 8);
		break;

	    default :
		fprintf (stderr,"psgl.insert_fb_cache(): garbage type %d?\n", type);
		return;

	    }	/* switch */
	}   /* while k < n */
}   /* insert_f_c */
#endif

/* insert generic graphics object into cache */
static int
insert_cache (
    int type,
    char *data,
    int n)
{
int k;
char *old_psgl_cache_base;
CACHE_ITEM *item;
POINT *vptr, *vbuff;
double xc, yc, zc;

    if (!psgl_backbuffer)		    /* if not drawing to back buffer, exit */
	return;

    /* are we about to overflow the item cache? */
    if (psgl_cache_index == psgl_max_items)    {
    	psgl_cache = (CACHE_ITEM *)realloc (psgl_cache, (psgl_max_items *= 2) * sizeof (CACHE_ITEM));
	if (psgl_cache == NULL)  {
	    fprintf (stderr,"psgl.insert_cache(): can't realloc() %d bytes for cache\n", 
		 psgl_max_items * sizeof (CACHE_ITEM));
	    exit (-1);
	    }
	fprintf (stderr,"psgl.insert_cache(): max cache items now %d (%d bytes)\n",
		    psgl_max_items, psgl_max_items * sizeof (CACHE_ITEM));
	}

    /* are we about to overflow the data space? */
    if (((long)psgl_cache_data + n * 3 * sizeof(float)) > ((long)psgl_cache_base + (psgl_max_bytes * 7) / 8))   {
	old_psgl_cache_base = psgl_cache_base;
	psgl_cache_base = (char *)realloc (psgl_cache_base, psgl_max_bytes *= 2);
	if (psgl_cache_base == NULL)  {
	    fprintf (stderr,"psgl.insert_cache(): can't realloc() %d bytes for psgl_cache_base\n", psgl_max_bytes);
	    exit (-1);
	    }
	fprintf (stderr,"psgl.insert_cache(): cache data size now %d bytes\n", psgl_max_bytes);
	for (k = 0; k < psgl_cache_index; k++)
	    psgl_cache[k].data = psgl_cache_base + (psgl_cache[k].data - old_psgl_cache_base);
	psgl_cache_data = psgl_cache_base + (psgl_cache_data - old_psgl_cache_base);
	}

    item = &psgl_cache[psgl_cache_index++];

    if (psgl_cache_index % 1000 == 0)
	fprintf (stderr, "psgl.insert_cache(): %d items.\n", psgl_cache_index);

    item->length = n;
    item->color = psgl_color;		    /* may be overwritten by readpixel() below */
    item->type = type;
    item->data = psgl_cache_data;	    /* point to next block of global free space */
    item->descriptor = PSGL_NORMAL;
    if (psgl_zbuffer)
        item->descriptor |= PSGL_ZBUFFER;
    item->timestamp = psgl_timestamp++;

    init_cache_bbox (item);

    if (type == CACHED_STRING)	{
        strncpy (item->data, data, n);		/* copy string to local space */
        psgl_cache_data += ((n + 3) / 4) * 4;	/* assure quad-word alignment */

	item->w = (float)strwidth(item->data);
	item->h = (float)getheight();		/* current char height in pixels */

#ifdef GETCPOS_WORKS
    	item->x = psgl_cposx + psgl_xorg;			/* globals */
    	item->y = psgl_cposy + psgl_yorg;
    	item->z = psgl_cz;
#else
    	item->x = psgl_cx;
    	item->y = psgl_cy;
    	item->z = psgl_cz;

	/* update */
	psgl_cx += item->w;
#endif
	item->xmin = item->x;
	item->ymin = item->y - getdescender();

	item->xmax = item->x + item->w;
	item->ymax = item->y + item->h - getdescender();
	}
    else    {
	vbuff = (POINT *)data;

        /* clear the pseudo-centroid */
	xc = yc = zc = 0.0;

        vptr = (POINT *)item->data;
        for (k = 0; k < n; k++) {		/* copy xformed vertices into cache */
	    vptr[k] = vbuff[k];

	    insert_cache_bbox (item, vptr[k].px, vptr[k].py, vptr[k].pz); 

	    /* compute pseudo-centroid */
	    xc += vptr[k].px;
	    yc += vptr[k].py;
	    zc += vptr[k].pz;
	    }

        /* store the pseudo-centroid of the object in cache */
        xc /= (float)n;
        yc /= (float)n;
        zc /= (float)n;

        item->x = xc;
        item->y = yc;
        item->z = zc;

	item->w = item->h = 0.0;	    /* ignored */

        /* bump the global allocation pointer */
        psgl_cache_data += (n * sizeof(float) * 3);
	}

    if (bbox_off_window(item))
	item->descriptor |= PSGL_OFFWINDOW;
}

/* GL REPLACEMENT ROUTINES */
/* implement all of the routines above as shadow copies */
/* convention: local routines start with capital letters */
/* note that all must be exported from module (i.e., non-static)

/* MATRIX OPS : just do gl matrix op, then getmatrix() to find result */

void 
Feedback(
    FB_TOKEN buffer[],
    long size)
{
    if (psgl_enabled) {
	fprintf (stderr, "ERROR (psgl): Can't use feedback mode while psgl routines enabled\n");
	return;
	}
    feedback (buffer, size);
}

long 
Endfeedback(
    FB_TOKEN buffer[])
{
    if (psgl_enabled) {
	fprintf (stderr, "ERROR (psgl): Can't use feedback mode while psgl routines enabled\n");
	return;
	}
    endfeedback (buffer);
}

void 
Pick(
    short buffer[], 
    long numnames)
{
    if (psgl_enabled)	{
	fprintf (stderr, "ERROR (psgl): Can't enter pick mode while psgl routines enabled\n");
	return;
	}

    pick (buffer, numnames);
}

long 
Endpick(
    short buffer[])
{
    if (psgl_enabled)	{
	fprintf (stderr, "ERROR (psgl): Can't affect pick mode while psgl routines enabled\n");
	return;
	}

    endpick (buffer);
}

void
Perspective(
    Angle fovy,
    float aspect,
    Coord near, Coord far)
{
    perspective(fovy, aspect, near, far);

    if (!psgl_enabled) return;

    psgl_cpnear = near;
    psgl_cpfar = far;

    psgl_ortho = FALSE;
}

void
Ortho(
    Coord left, Coord right, Coord bottom, Coord top, Coord near, Coord far)
{
    ortho(left, right, bottom, top, near, far);

    if (!psgl_enabled) return;

    psgl_cpnear = near;
    psgl_cpfar = far;

    psgl_ortho = TRUE;
}

void 
Ortho2(
    Coord left, Coord right, Coord bottom, Coord top)
{
    ortho2(left, right, bottom, top);

    if (!psgl_enabled) return;

    psgl_cpnear = -1.0;
    psgl_cpfar = 1.0;

    psgl_ortho = TRUE;
}

void
Viewport(
    Screencoord left, Screencoord right, Screencoord bottom, Screencoord top)
{
    viewport(left, right, bottom, top);

    if (!psgl_enabled) return;

    getviewport (&psgl_left, &psgl_right, &psgl_bottom, &psgl_top);    
}

void
Popviewport()
{
    popviewport();

    if (!psgl_enabled) return;

    getviewport (&psgl_left, &psgl_right, &psgl_bottom, &psgl_top);
}

void
Setdepth(
    Screencoord near, Screencoord far)
{
    setdepth(near, far);

    if (!psgl_enabled) return;

    psgl_zbnear = (long)near;
    psgl_zbfar  = (long)far;
}

void
Lsetdepth(
    long near, long far)
{
    lsetdepth(near, far);

    if (!psgl_enabled) return;

    psgl_zbnear = near;
    psgl_zbfar  = far;
}

void
Reshapeviewport()
{
    reshapeviewport();

    if (!psgl_enabled) return;

    getorigin (&psgl_xorg, &psgl_yorg);
    getsize   (&psgl_xsize, &psgl_ysize);

    getviewport (&psgl_left, &psgl_right, &psgl_bottom, &psgl_top);

    /* don't know how to get near & far... */
}

void 
Cpack(
    unsigned long pack)
{
    cpack (pack);
    psgl_color = pack;
}

void
rgbcolor(
    short red, short green, short blue)
{
    RGBcolor(red, green, blue);
    psgl_color = ((blue & 0xff) << 16) | ((green & 0xff) << 8) | (red & 0xff);
}

void
Arc(
    Coord x, Coord y, Coord radius,
    Angle startang, Angle endang)
{
    if (!psgl_enabled)	{
        arc(x, y, radius, startang, endang);
	return;
	}

    open_feedback();
    arc(x, y, radius, startang, endang);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Arcf(
    Coord x, Coord y, Coord radius,
    Angle startang, Angle endang)
{
    if (!psgl_enabled)	{
        arcf(x, y, radius, startang, endang);
	return;
	}

    open_feedback();
    arcf(x, y, radius, startang, endang);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Circ(
    Coord x, Coord y, Coord radius)
{
    if (!psgl_enabled)	{
        circ(x, y, radius);
	return;
	}

    open_feedback();
    circ(x, y, radius);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Circf(
    Coord x, Coord y, Coord radius)
{
    if (!psgl_enabled)	{
        circf(x, y, radius);
	return;
	}

    open_feedback();
    circf(x, y, radius);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Crv(
    Coord geom[4][3])
{
    if (!psgl_enabled)	{
        crv (geom);
	return;
	}

    open_feedback();
    crv (geom);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Crvn(
    long n,
    Coord geom[][3])
{
    if (!psgl_enabled)	{
	crvn (n, geom);
	return;
	}

    open_feedback();
    crvn (n, geom);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Rcrv(
    Coord geom[4][4])
{
    if (!psgl_enabled)	{
        rcrv (geom);
	return;
	}

    open_feedback();
    rcrv (geom);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Rcrvn(
    long n,
    Coord geom[][4])
{
    if (!psgl_enabled)	{
        rcrvn (n, geom);
	return;
	}

    open_feedback();
    rcrvn (n, geom);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Patch(
    Matrix geomx, Matrix geomy, Matrix geomz)
{
    if (!psgl_enabled)	{
        patch (geomx, geomy, geomz);
	return;
	}

    open_feedback();
    patch (geomx, geomy, geomz);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Rpatch(
    Matrix geomx, Matrix geomy, Matrix geomz, Matrix geomw)
{
    if (!psgl_enabled)	{
        rpatch (geomx, geomy, geomz, geomw);
	return;
	}

    open_feedback();
    rpatch (geomx, geomy, geomz, geomw);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Cmov(
    Coord x, Coord y, Coord z)
{
    if (!psgl_enabled)	{
        cmov(x, y, z);
	return;
	}

    open_feedback();
    cmov (x, y, z);
    insert_feedback_cache (psgl_fbbuff, close_feedback());    
}

void
Charstr(
    String str)
{
    if (!psgl_enabled)	{
	charstr(str);
	return;
	}

#ifdef GETCPOS_WORKS
    /*
    getcpos() is brain-dead.  it only returns x,y.
    x,y,z are grabbed in Cmov(), above.
    */
    getcpos (&psgl_cposx, &psgl_cposy);		/* globals */
#endif

    charstr(str);

    if (!psgl_enabled) return;

    /* cache the string at current _screen_ position */
    insert_cache (CACHED_STRING, str, strlen (str) + 1);	    /* don't forget null-terminator */
}

/* in postscript, points are just zero-length lines */
void
Pnt(
    Coord x, Coord y, Coord z)
{
    if (!psgl_enabled)	{
	pnt (x, y, z);
	return;
	}

    open_feedback();
    pnt (x, y, z);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Draw(
    Coord x, Coord y, Coord z)
{
    if (!psgl_enabled)	{
        draw (x, y, z);
	return;
	}

    open_feedback();
    draw (x, y, z);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void Rdr(
    Coord x, Coord y, Coord z)
{
    
    if (!psgl_enabled)	{
        rdr (x, y, z);
	return;
	}

    open_feedback();
    rdr (x, y, z);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Rect(
    Coord x1, Coord y1, Coord x2, Coord y2)
{
    if (!psgl_enabled)	{
        rect(x1,y1,x2,y2);
	return;
	}

    open_feedback();
    rect(x1,y1,x2,y2);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Rectf(
    Coord x1, Coord y1, Coord x2, Coord y2)
{
    if (!psgl_enabled)	{
        rectf(x1,y1,x2,y2);
	return;
	}

    open_feedback();
    rectf(x1,y1,x2,y2);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polf(
    long n,
    Coord parray[][3])
{
    if (!psgl_enabled)	{
        polf (n, parray);
	return;
	}

    open_feedback();
    polf (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polfi(
    long n,
    Icoord parray[][3])
{
    if (!psgl_enabled)	{
        polfi (n, parray);
	return;
	}

    open_feedback();
    polfi (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polfs(
    long n,
    Scoord parray[][3])
{
    if (!psgl_enabled)	{
        polfs (n, parray);
	return;
	}

    open_feedback();
    polfs (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polf2(
    long n,
    Coord parray[][2])
{
    if (!psgl_enabled)	{
        polf2 (n, parray);
	return;
	}

    open_feedback();
    polf2 (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polf2i(
    long n,
    Icoord parray[][2])
{
    if (!psgl_enabled)	{
        polf2i (n, parray);
	return;
	}

    open_feedback();
    polf2i (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Polf2s(
    long n,
    Scoord parray[][2])
{
    if (!psgl_enabled)	{
        polf2s (n, parray);
	return;
	}

    open_feedback();
    polf2s (n, parray);
    insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnpolygon()
{
    if (psgl_enabled)
        open_feedback();
    bgnpolygon();
}

void
Endpolygon()
{
    endpolygon();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnline()
{
    if (psgl_enabled)
        open_feedback();
    bgnline();
}

void
Endline()
{
    endline();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnpoint()
{
    if (psgl_enabled)
        open_feedback();
    bgnpoint();
}

void
Endpoint()
{
    endpoint();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnclosedline()
{
    if (psgl_enabled)
        open_feedback();
    bgnclosedline();
}

void
Endclosedline()
{
    endclosedline();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgntmesh()
{
    if (psgl_enabled)
        open_feedback();
    bgntmesh();
}

void
Endtmesh()
{
    endtmesh();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnqstrip()
{
    if (psgl_enabled)
        open_feedback();
    bgnqstrip();
}

void
Endqstrip()
{
    endqstrip();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Zbuffer(
    Boolean bool)
{
    if (!psgl_enabled)	{
	zbuffer(bool);
	return;
        }

    psgl_zbuffer = bool;

    zbuffer (bool);
}

void
Backface(
    Boolean b)
{
    if (!psgl_enabled)	{
	backface(b);
	return;
	}
    /* do nothing; we're forcing backface to a fixed value based on trigger key sequence */
}

long
Winopen(
    String name)
{
char *rname;

    if ((rname = (char *)rindex (name, (int)'/')) != NULL)
	name = (String)rname+1;
    
    strcpy (psgl_prefix, name);
    return (winopen(name));
}

/* this is probably the source of the worst performance hit */
long
Qread(
    short *data)
{
long rval;
static short shift_pressed = FALSE;
static short left_control_pressed = FALSE;
static short right_control_pressed = FALSE;

    rval = qread(data);

    switch (rval)   {
	case LEFTCTRLKEY :
	    left_control_pressed = *data;
	    break;
	case RIGHTCTRLKEY : 
	    right_control_pressed = *data;
	    break;
	case LEFTSHIFTKEY :
	case RIGHTSHIFTKEY :
	    shift_pressed = *data;
	    break;
	default:
	    if (left_control_pressed && shift_pressed)	{
		if ((rval == PRINTSCREENKEY) && (!psgl_triggered) && *data)    {
		    psgl_triggered = TRUE;				   /* won't have effect until next swap */
		    psgl_userbackface = FALSE;
		    qenter (REDRAW, 1);
		    }
		else if ((rval == PKEY) && *data)   {
		    psgl_mode = PSGL_POSTSCRIPT;	
		    fprintf (stderr, "psgl.Qread(): postscript output mode selected.\n");
		    }
		else if ((rval == SKEY) && *data)   {
		    psgl_mode = PSGL_SHOWCASE;	
		    fprintf (stderr, "psgl.Qread(): showcase output mode selected.\n");
		    }
		}
	    else if (right_control_pressed && shift_pressed && (!psgl_triggered) && (rval == PRINTSCREENKEY) && *data)    {
		psgl_triggered = TRUE;	    /* won't have effect until next swap */
		psgl_userbackface = TRUE;
		qenter (REDRAW, 1);
		}
	    break;
	}

    return (rval);
}

void
Bgntrim()
{
    if (psgl_enabled)	
	open_feedback();
    bgntrim();
}

void
Endtrim()
{
    endtrim();
    
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Bgnsurface()
{
    if (psgl_enabled)	
	open_feedback();
    bgnsurface();
}

void
Endsurface()
{
    endsurface();
    if (psgl_enabled)
        insert_feedback_cache (psgl_fbbuff, close_feedback());
}

void
Clear()
{
    clear();

    if (!psgl_enabled) return;
/* can't do this stuff... what if only clearing a viewport?
    psgl_timestamp = 0;
    psgl_cache_index = 0;
    psgl_cache_data = psgl_cache_base;
    psgl_cache[0].data = NULL;
    psgl_fb_omitted = FALSE;
    psgl_timestamp = 0;
    psgl_equal_z = 0;
*/
}

void
Backbuffer(
    Boolean b)
{
    backbuffer(b);

    psgl_backbuffer = b;
}

void
Frontbuffer(
    Boolean b)
{
    frontbuffer(b);

    psgl_frontbuffer = b;
}

/* postscript ops */
#define PPI 72.0
#define XMARGIN (0.25 * PPI)
#define YMARGIN (0.25 * PPI)
#define PAGE_WIDTH ((8.5 * PPI) - 2 * XMARGIN)
#define PAGE_HEIGHT ((11.0 * PPI) - 2 * YMARGIN)

/* map coords [0..psgl_xsize, 0..psgl_ysize] to [XMARGIN..XMARGIN+PAGE_WIDTH, YMARGIN..YMARGIN+PAGE_HEIGHT] */
static	float psgl_sx, psgl_sy, psgl_xmargin, psgl_ymargin;
#define PSGLX(x) (psgl_xmargin + (psgl_sx * x))
#define PSGLY(y) (psgl_ymargin + (psgl_sy * y))

static void
psgl_derive_coords ()
{
float hw;
float page_width, page_height;

    if (psgl_mode == PSGL_POSTSCRIPT)	{
	psgl_xmargin = XMARGIN;
	psgl_ymargin = YMARGIN;
	page_width = PAGE_WIDTH;
	page_height = PAGE_HEIGHT;
	}
    else    {	    /* NYI */
	psgl_xmargin = 0.0;
	psgl_ymargin = 0.0;
	page_width = 8.5;
	page_height = 11;
        }

    /* an explanation of coordinates used.  */
    /* all cache entries are stored in (floating point) window coords:
	[0.0..psgl_xsize] x [0.0..psgl_ysize] x [psgl_zbnear .. psgl_zbfar]	*/
    /* also, widths, heights, and depths are in these coordinates.  */

    /* we map this rectangle to a postscript page, preserving aspect ratio */
    /* the postscript page is in points, 72/inch; 8 1/2" wide x 11" tall */

    /* find the height/width ratio of current window */
    hw = (float)psgl_ysize / (float)psgl_xsize;

    if (hw > page_height/page_width)	{	/* scale height to fit */
	psgl_sy = page_height / (float)psgl_ysize;
	psgl_sx = psgl_sy;
	}
    else    {
	psgl_sx = page_width / (float)psgl_xsize;
	psgl_sy = psgl_sx;
	}
}

static void
psgl_derive_bbox()
{
int k;
CACHE_ITEM *item;

    /* compute bounding box of bounding boxes; don't include offwindow */
    psgl_xmin = psgl_ymin =  1E6;
    psgl_xmax = psgl_ymax = -1E6;
    for (k = 0; k < psgl_cache_index; k++)	{
	item = &psgl_cache[k];
	if (!(item->descriptor & PSGL_OFFWINDOW))   {
	    insert_bbox (item->xmin, item->ymin);
	    insert_bbox (item->xmax, item->ymin);
	    insert_bbox (item->xmax, item->ymax);
	    insert_bbox (item->xmin, item->ymax);
	    }
	}   /* for */

    /* clip bounding box to graphics window */
    psgl_xmin = MAX (0.0, psgl_xmin);
    psgl_ymin = MAX (0.0, psgl_ymin);
    psgl_xmax = MIN (psgl_xmax, psgl_xsize);
    psgl_ymax = MIN (psgl_ymax, psgl_ysize);

    /* convert to postscript coords */
    psgl_xmin = PSGLX (psgl_xmin);
    psgl_ymin = PSGLY (psgl_ymin);

    psgl_xmax = PSGLX (psgl_xmax);
    psgl_ymax = PSGLY (psgl_ymax);

    /* expand by one percent for border objects */
    psgl_w = psgl_xmax - psgl_xmin;
    psgl_h = psgl_ymax - psgl_ymin;

    psgl_xmin -= psgl_w * 0.0125;
    psgl_xmax += psgl_w * 0.0125;

    psgl_ymin -= psgl_h * 0.0125;
    psgl_ymax += psgl_h * 0.0125;
}

static char moveto[] = "moveto";
static char lineto[] = "lineto";
static char rlineto[] = "rlineto";
static char newpath[] = "newpath";
static char closepath[] = "closepath";
static char stroke[] = "stroke";
static char fill[] = "fill";
static char setrgbcolor[] = "setrgbcolor";

static void
psgl_emit_prologue()
{
  fprintf  (stderr, "psgl.Swapbuffers(): writing '%s'\n", psgl_fname);

  if (psgl_mode == PSGL_POSTSCRIPT) {
    /* write EPS header */
    fprintf (psgl_file, "%%!\n");
    fprintf (psgl_file, "%%%%Title: (%s)\n", psgl_fname);
    fprintf (psgl_file, "%%%%Creator: psgl (sgi gl to postscript conversion)\n");
    fprintf (psgl_file, "%%%%BoundingBox: %4.2f %4.2f %4.2f %4.2f\n", psgl_xmin, psgl_ymin, psgl_xmax, psgl_ymax);
    fprintf (psgl_file, "%%%%DocumentFonts: Times-Roman Times-Italic Times-Bold Courier-Bold\n");
    fprintf (psgl_file, "%%%%Pages: 1\n");

#ifdef PSVIEW_WORKS
    fprintf (psgl_file, "/psgldict 8 dict def\n");	/* new 8-item dictionary */
    fprintf (psgl_file, "psgldict begin\n");	/* push new dictionary */
    /* special cases : points, triangles, quadrilaterals */
    /* pnt takes x,y on stack; duplicates top two args */
    fprintf (psgl_file, "/pnt {newpath moveto lineto closepath stroke } def\n");
#endif

    fprintf (psgl_file, "/mt {moveto} def\n");	    /* define mt == moveto */
    fprintf (psgl_file, "/lt {lineto} def\n");	    /* define lt == lineto */
    fprintf (psgl_file, "/rl {rlineto} def\n");	    /* define rl == rlineto */
    fprintf (psgl_file, "/np {newpath} def\n");	    /* define np == newpath */
    fprintf (psgl_file, "/cp {closepath} def\n");   /* define cp == closepath */
    fprintf (psgl_file, "/st {stroke} def\n");	    /* define st == stroke */
    fprintf (psgl_file, "/fi {fill} def\n");	    /* define fi == fill */
    fprintf (psgl_file, "/sc {setrgbcolor} def\n"); /* define sc == setrgbcolor */

    sprintf (moveto,"mt");
    sprintf (lineto,"lt");
    sprintf (rlineto,"rt");
    sprintf (newpath,"np");
    sprintf (closepath,"cp");
    sprintf (stroke,"st");
    sprintf (fill,"fi");
    sprintf (setrgbcolor,"sc");

    fprintf (psgl_file, "\ngsave\n\n");
    /* invoke defaults */
    fprintf (psgl_file, "1 setlinecap\n");		    /* round ends */
    fprintf (psgl_file, "1.0 setlinewidth\n");
    fprintf (psgl_file, "/Times-Bold findfont 18 scalefont setfont\n");

    /* clip to bounding box */
    fprintf (psgl_file, "newpath\n");
    fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmin, psgl_ymin, moveto);
    fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmax, psgl_ymin, lineto);
    fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmax, psgl_ymax, lineto);
    fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmin, psgl_ymax, lineto);
    fprintf (psgl_file, "closepath clip\n");

    /* show the clip path */
    if (psgl_debug_level)  {
        fprintf (psgl_file, "[2 4] 0 setdash 3.0 setlinewidth\n");
        fprintf (psgl_file, "newpath\n");
        fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmin, psgl_ymin, moveto);
        fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmax, psgl_ymin, lineto);
        fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmax, psgl_ymax, lineto);
        fprintf (psgl_file, "%4.2f %4.2f %s\n", psgl_xmin, psgl_ymax, lineto);
        fprintf (psgl_file, "closepath stroke\n");
        fprintf (psgl_file, "[] 0 setdash 1.0 setlinewidth\n");
	}
  }
  else	{
    fprintf (psgl_file, "sungremlinfile\n");
    fprintf (psgl_file, "1 0.00 0.00\n");
    }    
}

static void
psgl_order_cache()
{
    fprintf  (stderr, "psgl.Swapbuffers(): sorting %d items... ", psgl_cache_index);
    /* sort the cache by z depth, time drawn, etc. */
    qsort ((char *)psgl_cache, psgl_cache_index, sizeof (CACHE_ITEM), cicomp);
    fprintf  (stderr, "done.\n");
}

static void
psgl_emit_cache()
{
CACHE_ITEM *item;
int type, i, k;
POINT *vptr; 
static long current_color = 0L;
static float current_fontsize = -1.0;

    current_color = 0L;
    current_fontsize = -1.0;

    for (k = 0; k < psgl_cache_index; k++)  {

	item = &psgl_cache[k];
	type = item->type;

	if (psgl_mode == PSGL_POSTSCRIPT)   {
	    if (psgl_debug_level)
	      fprintf (psgl_file, "%%%% psgl obj %d%s%s (z %4.2f, zbuf %d)\n", item->timestamp,
		item->descriptor & PSGL_NOCOLOR ? " (no color) " : "",
		item->descriptor & PSGL_OFFWINDOW ? " (offwindow) " : "", item->z, item->descriptor & PSGL_ZBUFFER);

	    if (k == 0 || item->color != current_color) {
		current_color = item->color;
		fprintf (psgl_file, "%4.2f %4.2f %4.2f %s\n",
		    ((current_color & 0x000000FF)      ) / 255.0,	    /* red	*/
		    ((current_color & 0x0000FF00) >>  8) / 255.0,	    /* green	*/
		    ((current_color & 0x00FF0000) >> 16) / 255.0,	    /* blue	*/
		    setrgbcolor);
		}
	    }

#ifdef BBOX
	if (psgl_mode == PSGL_POSTSCRIPT)	{
            /* show object bbox and id */
	    if (psgl_debug_level)  {
		fprintf (psgl_file, "0.0 0.0 0.0 %s\n", setrgbcolor);

		fprintf (psgl_file, "[2 4] 0 setdash 1.0 setlinewidth\n");
		fprintf (psgl_file, "newpath\n");
		fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->xmin), PSGLY(item->ymin), moveto);
		fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->xmax), PSGLY(item->ymin), lineto);
		fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->xmax), PSGLY(item->ymax), lineto);
		fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->xmin), PSGLY(item->ymax), lineto);
		fprintf (psgl_file, "closepath stroke\n");
		fprintf (psgl_file, "[] 0 setdash 1.0 setlinewidth\n");

		fprintf (psgl_file, "%4.2f %4.2f %4.2f %s\n",
		    ((current_color & 0x000000FF)      ) / 255.0,	    /* red	*/
		    ((current_color & 0x0000FF00) >>  8) / 255.0,	    /* green	*/
		    ((current_color & 0x00FF0000) >> 16) / 255.0,	    /* blue	*/
		    setrgbcolor);
		}
	    }	    /* POSTSCRIPT */
	else	{   /* SHOWCASE */
	    }
#endif

	switch (type)  {
	    case CACHED_POINT :
				vptr = (POINT *)item->data;
				if (psgl_mode == PSGL_POSTSCRIPT)	{
				    for (i = 0; i < item->length; i++)	{
					fprintf (psgl_file, "%s\n", newpath);
					fprintf (psgl_file, "%4.2f %4.2f %s %4.2f %4.2f %s\n",
					    PSGLX(vptr[i].px), PSGLY(vptr[i].py), moveto, 
					    PSGLX(vptr[i].px), PSGLY(vptr[i].py), lineto);
					fprintf (psgl_file, "%s %s\n", closepath, stroke);
					}
				    }
				else	{   /* SHOWCASE */
				    }
				break;

	    case CACHED_VECTOR :
	    case CACHED_VECTORLOOP :
	    case CACHED_POLYGON :
				vptr = (POINT *)item->data;
				if (psgl_mode == PSGL_POSTSCRIPT)	{
				    fprintf (psgl_file, "%s\n", newpath);

				    for (i = 0; i < item->length; i++)
					fprintf (psgl_file, "%4.2f %4.2f %s\n",
					    PSGLX(vptr[i].px), PSGLY(vptr[i].py), i ? lineto : moveto);
				    if (type == CACHED_VECTORLOOP || type == CACHED_POLYGON)
					fprintf (psgl_file, "%s\n", closepath);
				    if (item->type == CACHED_POLYGON)
					fprintf (psgl_file, "%s\n", fill);
				    else
					fprintf (psgl_file, "%s\n", stroke);
				    }
				else	{   /* SHOWCASE */
				    if (type == CACHED_POLYGON)
					fprintf (psgl_file, "POLYGON\n");
				    else
					fprintf (psgl_file, "VECTOR\n");
				    for (i = 0; i < item->length; i++)
					fprintf (psgl_file, "%4.2f %4.2f\n", PSGLX(vptr[i].px), PSGLY(vptr[i].py));
				    if (type != CACHED_VECTOR)
					fprintf (psgl_file, "%4.2f %4.2f\n", PSGLX(vptr[0].px), PSGLY(vptr[0].py));
				    fprintf (psgl_file, "*\n");
				    if (type != CACHED_POLYGON)
					fprintf (psgl_file, "5 1\n");
				    else    {       /* filled poly 0 = white, 64 = black */
					int grn_color;

					grn_color = (item->color & 0x000000FF) 
						+   ((item->color & 0x0000FF00) >>  8)
						+   ((item->color & 0x00FF0000) >> 16);
					grn_color = ((grn_color / (3.0 * 255.0)) * 16.0);
					grn_color = 16 - grn_color;
					fprintf (psgl_file, "0 %d\n", grn_color);
					}
				    fprintf (psgl_file, "0 \n");
				    }
				break;

	    case CACHED_STRING :
				if (psgl_mode == PSGL_POSTSCRIPT)	{
				    /* convert char height (in pixels) to postscript height (in points) */
				    item->h *= psgl_sy;		/* convert to points */
				    if (item->h != current_fontsize)
					fprintf (psgl_file, "/Times-Bold findfont %4.2f scalefont setfont\n",
					    current_fontsize = item->h);
				    fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->x), PSGLY(item->y), moveto);
				    fprintf (psgl_file, "(%s) show\n", item->data);
				    }
				else	{   /* SHOWCASE */
				    fprintf (psgl_file, "BOTLEFT\n");
				    fprintf (psgl_file, "%4.2f %4.2f\n", PSGLX(item->x), PSGLY(item->y));
				    fprintf (psgl_file, "*\n");
				    fprintf (psgl_file, "3 2\n");   /* bold, size 2 */
				    fprintf (psgl_file, "%d %s\n", item->length - 1, item->data);
				    }
				break;
	    }	/* switch */

	if (psgl_mode == PSGL_POSTSCRIPT)	{
	    /* show object id */
	    if (psgl_debug_level)  {
		char id[12];

		fprintf (psgl_file, "0.0 0.0 0.0 %s\n", setrgbcolor);

		fprintf (psgl_file, "%4.2f %4.2f %s\n", PSGLX(item->x), PSGLY(item->y), moveto);
		sprintf (id, "%d", item->timestamp);
		fprintf (psgl_file, "(%s) show\n", id);

		fprintf (psgl_file, "%4.2f %4.2f %4.2f %s\n",
		    ((current_color & 0x000000FF)      ) / 255.0,	    /* red	*/
		    ((current_color & 0x0000FF00) >>  8) / 255.0,	    /* green	*/
		    ((current_color & 0x00FF0000) >> 16) / 255.0,	    /* blue	*/
		    setrgbcolor);
		}
	    }	/* POSTSCRIPT */
	}
}

static void
psgl_emit_epilogue()
{
    if (psgl_mode == PSGL_POSTSCRIPT) {
#ifdef PSVIEW_WORKS
	fprintf (psgl_file, "end %% psgldict\n");	/* pop local dict */
#endif
        fprintf (psgl_file, "\ngrestore\n");

	if (psgl_epilogue != NULL)
	    fprintf (psgl_file, psgl_epilogue);
        }	/* POSTSCRIPT */
    else    {
	fprintf (psgl_file, "-1\n");
	}	/* SHOWCASE */

    fprintf  (stderr, "psgl.psgl_emit_epilogue(): closing '%s'\n", psgl_fname);
    fclose (psgl_file);
    psgl_file = NULL;

    fprintf  (stderr, "psgl.Swapbuffers(): wrote %d objects\n", psgl_cache_index);
    if (psgl_equal_z)
        fprintf  (stderr, "psgl.Swapbuffers(): there were %d z collisions in qsort()\n", psgl_equal_z);
    if (psgl_fb_omitted)
        fprintf  (stderr, "psgl.Swapbuffers(): feedback buffer possibly insufficient this run.\n");
}

void
Swapbuffers()
{
    swapbuffers();

    psgl_install();

    if (psgl_triggered)	{
	psgl_triggered = FALSE;
	(void)psgl_open(NULL);
	return;
	}

    if (!psgl_enabled) return;

    psgl_enabled = FALSE;

    if (!psgl_cache_index)  {
	fprintf (stderr, "psgl.Swapbuffers(): 0 items in cache\n");
        psgl_restore_client_state();
	return;
	}

    psgl_derive_coords ();

    psgl_derive_bbox();

    psgl_order_cache();

    psgl_emit_prologue ();
 
    psgl_emit_cache();

    psgl_emit_epilogue();

    psgl_restore_client_state();
    
    /* force client to redraw ONCE; this will restore their window to its predump state */
    qenter (REDRAW, 1);
}
